import datetime as dt
from typing import Optional

from pydantic import BaseModel


class PeriodTime(BaseModel):
    start_time: dt.time  # 開始時間
    end_time: dt.time    # 結束時間


class CourseInfoSchedule(BaseModel):
    day_of_week: str                # 星期
    period: str                     # 節次
    period_times: list[PeriodTime]  # 節次時間


class CourseInfo(BaseModel):
    """ sso 學期選課資料 """
    course_no: str                         # 學期課號 Course Serial No.
    curriculum_no: str                     # 系所課號 Curriculum No.
    name: str                              # 課程名稱 Course Name
    class_name: str                        # 開課班級 Class
    team: str | None                       # 班別 Team
    req_elect: str                         # 修別 Required/Elective
    credits: str                           # 學分組合 Credits 講授－實習－學分數
    schedule: CourseInfoSchedule | None    # 星期-節次 Schedule
    location: str | None                   # 教室 Location
    instructors: list[str]                 # 授課教師 Instructor
    selected: int | None                   # 修課人數 Sel.
    max_capacity: int | None               # 人數限制 Max
    remarks: Optional[str] = None          # 備註 Remarks
    dist_url: Optional[str] = None         # 遠距上課網址 Distance Learning
    materials_url: Optional[str] = None    # 教材網站 Teaching Materials Website



from datetime import time

import pytz

TZ = pytz.timezone("Asia/Taipei")  # UTC+8
CODE_SEQUENCE = "ABCDYEFGHZ"
A_HOUR = 8


def get_period_time(code: str):
    if len(code) != 1:
        raise ValueError(f"Invalid period code: {code}")

    if code not in CODE_SEQUENCE:
        raise ValueError(f"Invalid period code: {code}")

    idx = CODE_SEQUENCE.index(code)

    # 計算開始和結束時間
    start_time = time(A_HOUR + idx, 10, tzinfo=TZ)
    end_time = time(A_HOUR + idx + 1, 0, tzinfo=TZ)

    return start_time, end_time



import base64
import re
from typing import Optional

import requests
from bs4 import BeautifulSoup

from ._types import CourseInfo, CourseInfoSchedule, PeriodTime
from .errors import *
from .tronclass import TronClassClient
from .utils import get_period_time

LOGIN_URL = "https://webapp.yuntech.edu.tw/YunTechSSO/Account/Login"  # SSO 登入 API URL
CAPTCHA_URL = "https://webapp.yuntech.edu.tw/YunTechSSO/Captcha/Number"  # 驗證碼圖片 URL
ECLASS_LOGIN_URL = "https://eclass.yuntech.edu.tw/yuntech/sso-login"  # eclass 登入 URL


class NYUSTSSOClient:
    def __init__(self) -> None:
        self.session = requests.Session()

        # 是否登入過 WebNewCAS
        self.__login_WebNewCAS = False

    def __get_login_data(self, username: str, password: str, captcha: str) -> dict:
        response = self.session.get(LOGIN_URL)
        soup = BeautifulSoup(response.text, 'html.parser')
        return {
            "__RequestVerificationToken": soup.find('input', {'name': '__RequestVerificationToken'}).get('value'),
            "RedirectTo": None,
            "auth_appid": None,
            "auth_token": None,
            "auth_sig": None,
            "auth_ts": None,
            "auth_nonce": None,
            "auth_version": None,
            "pRememberMe": False,
            "pLoginName": username,
            "pLoginPassword": password,
            "pSecretString": captcha
        }

    def __handle_miraculous_redirect(self, url: str) -> str:
        """ 神奇的重定向處理 偉哉 Yuntech sso """
        r = self.session.get(url)
        soup = BeautifulSoup(r.text, 'html.parser')
        redirect_url = soup.find('a').get('href')
        return redirect_url

    def fetch_captcha(self) -> bytes:
        """ 獲取 sso 登入頁面的驗證碼 """
        captcha = self.session.get(CAPTCHA_URL)
        return base64.b64decode(captcha.text)

    def login(self, username: str, password: str, captcha: str) -> None:
        """ 登入 NYUST SSO & TronClass """

        # prepare login data
        login_data = self.__get_login_data(username, password, captcha)

        # login request
        login_resp = self.session.post(LOGIN_URL, data=login_data)
        # with open("login.html", "w", encoding="utf8") as f:
        #     f.write(login_resp.text)

        # 錯誤驗證 - 偉哉 Yuntech sso 錯誤時回傳 200
        if "Account not exist or registered" in login_resp.text:
            raise AccountNotRegisteredError()  # 帳號不存在或未註冊

        if "Invalid validation code" in login_resp.text:
            raise CaptchaError()  # 驗證碼錯誤

        # 偉哉 Yuntech sso 重定向不用 302
        # 登入 eclass -> 重定向到 sso -> 重定向到 eclass
        redirect_url = self.__handle_miraculous_redirect(ECLASS_LOGIN_URL)
        self.session.get(redirect_url)

    def get_tronclass_client(self):
        return TronClassClient(self.session)

    def fetch_announcements(self) -> dict:
        """ 獲取公告(?) (這個好像會是空的) """
        response = self.session.get("https://eclass.yuntech.edu.tw/api/announcements")
        return response.json()

    def fetch_todos(self) -> dict:
        """ 獲取待辦事項 """
        response = self.session.get("https://eclass.yuntech.edu.tw/api/todos")
        return response.json()

    def fetch_bulletins_latest(self) -> dict:
        """ 獲取最新公告 """
        response = self.session.get("https://eclass.yuntech.edu.tw/api/bulletins/latest")
        return response.json()

    def fetch_forum_topic_categories(self, course_id: str) -> dict:
        """ 獲取課程中所有討論區 """
        response = self.session.get(f"https://eclass.yuntech.edu.tw/api/courses/{course_id}/topic-categories")
        return response.json()

    def fetch_forum_category(self, category_id: str, page_size: int = 10):
        """ 獲取課程中某個討論區中的所有文章 """
        # TODO 好像還可以帶一堆參數
        response = self.session.get(f"https://eclass.yuntech.edu.tw/api/forum/categories/{category_id}?page_size={page_size}")
        return response.json()

    def fetch_topic(self, topic_id: str):
        """ 獲取討論串中某個文章的內容 """
        # TODO 好像還可以帶一堆參數
        response = self.session.get(f"https://eclass.yuntech.edu.tw/api/topics/{topic_id}")
        return response.json()

    def fetch_enrollments(self, course_id: str):
        response = self.session.get(f'https://eclass.yuntech.edu.tw/api/course/{course_id}/enrollments')
        return response.json()

    def fetch_my_courses(self, year: Optional[int] = None, semester: Optional[int] = None) -> list[CourseInfo]:
        if semester not in [None, 1, 2]:  # 1: 上學期, 2: 下學期
            raise ValueError("semester should be 1 or 2")

        if (year is None) != (semester is None):  # 都設置或都不設置
            raise ValueError("year and semester should be both None or both not None")

        api_url = "https://webapp.yuntech.edu.tw/WebNewCAS/StudentFile/Course/"

        # 獲取當前學期課程資訊
        if not self.__login_WebNewCAS:  # 如果還沒登入過 WebNewCAS -> 處理 sso 重定向
            self.__login_WebNewCAS = True
            redirect_url = self.__handle_miraculous_redirect(api_url)
            result_resp = self.session.get(redirect_url)
        else:
            result_resp = self.session.get(api_url)

        if year != None and semester != None:  # 如果有指定學期 -> 重新發送請求 (偉哉 Yuntech sso)
            if len(str(year)) != 3:  # 3-digit number
                raise ValueError("year should be a 3-digit number")

            soup = BeautifulSoup(result_resp.text, 'html.parser')
            with open("get.html", "w", encoding="utf8") as f:
                f.write(result_resp.text)

            post_data = {
                "__EVENTTARGET": "ctl00$MainContent$AcadSeme",
                "__EVENTARGUMENT": None,
                "__LASTFOCUS": None,
                "__VIEWSTATE": soup.find('input', {'name': '__VIEWSTATE'}).get('value'),
                "__VIEWSTATEGENERATOR": soup.find('input', {'name': '__VIEWSTATEGENERATOR'}).get('value'),
                "__VIEWSTATEENCRYPTED": "",
                "__EVENTVALIDATION": soup.find('input', {'name': '__EVENTVALIDATION'}).get('value'),
                "ctl00$MainContent$AcadSeme": f"{year}{semester}"
            }
            result_resp = self.session.post(api_url, data=post_data)

        data = []
        soup = BeautifulSoup(result_resp.text, 'html.parser')
        table = soup.find('table', {"id": "ctl00_MainContent_StudCour_GridView"})
        rows = table.find_all('tr')
        for row in rows:
            cells = row.find_all('td')  # find all cells
            row_data = [cell.get_text(strip=True) for cell in cells]  # get text from each cell

            if len(row_data) == 0:
                continue

            # extract_max_capacity
            match = re.search(r'限(\d+)人', row_data[10])
            max_capacity = match.group(1) if match else None

            # split schedule and location
            match = re.match(r'(\d+)-([A-Z]+)/(.*)', row_data[7])
            if match:
                period = match.group(2)        # 節次
                location = match.group(3)      # 教室
                schedule = CourseInfoSchedule(
                    day_of_week=match.group(1),
                    period=period,
                    period_times=[PeriodTime(start_time=s_time, end_time=e_time) for p in period for s_time, e_time in [get_period_time(p)]]
                )

            data.append(CourseInfo(
                course_no=row_data[0],
                curriculum_no=row_data[1],
                name=row_data[2],
                class_name=row_data[3],
                team=row_data[4],
                req_elect=row_data[5],
                credits=row_data[6],
                schedule=schedule,
                location=location,
                instructors=row_data[8].split(","),
                selected=int(row_data[9]),
                max_capacity=max_capacity,
                remarks=row_data[11],
                dist_url=row_data[12],
                materials_url=row_data[13]
            ))
        return data

    def comment(self, reply_id: str, content: str):
        """ 回覆討論串中的留言 """
        payload = {"content": f"<p> {content} </p>", "uploads": []}
        r = self.session.post(f"https://eclass.yuntech.edu.tw/api/replies/{reply_id}/comments", json=payload)
        return r



import time
from typing import TYPE_CHECKING, Literal, Union

from pydantic import BaseModel
from requests import Session

from .base import BaseResourceModel

if TYPE_CHECKING:
    from .course import Course


class Video(BaseModel):
    duration: float
    id: int
    resolution: str


class Upload(BaseModel):
    id: int
    name: str
    videos: list[Video]


class Activity(BaseResourceModel):
    id: int
    type: Literal['online_video', 'material'] | str
    title: str
    uploads: list[Upload]
    module_id: int
    syllabus_id: int

    @property
    def course(self) -> "Course":
        if self._course is None:
            pass  # TODO fetch course
        return self._course  # type: ignore

    def __init__(self, *, session: Session, course: "Course", **data):
        super().__init__(session=session, **data)
        self._course = course

    def _complete_video(self):
        def _read_chunk(start: int, end: int):
            global_data = self.course.global_data

            rsp = self._session.post(
                f"https://eclass.yuntech.edu.tw/api/course/activities-read/{self.id}",
                json={"start": start, "end": end}
            )
            if rsp.status_code != 201:
                raise ValueError(f"Failed to read video chunk: {rsp.json()}")

            rsp = self._session.post(
                "https://eclass.yuntech.edu.tw/statistics/api/online-videos",
                json={
                    "user_id": global_data['user']['id'],
                    "org_id": global_data['course']['orgId'],
                    "course_id": global_data['course']['id'],
                    "module_id": self.module_id,
                    "syllabus_id": self.syllabus_id,
                    "activity_id": self.id,
                    "upload_id": 0 if len(self.uploads) == 0 else self.uploads[0].id,
                    "reply_id": None,
                    "comment_id": None,
                    "forum_type": "",
                    "action_type": "play",
                    "is_teacher": False,
                    "is_student": True,
                    "ts": int(time.time() * 1000),
                    "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36 Edg/131.0.0.0",
                    "meeting_type": "online_video",
                    "start_at": start,
                    "end_at": end,
                    "duration": end - start,
                    "org_name": global_data['course']['orgName'],
                    "org_code": global_data['course']['orgCode'],
                    "user_no": global_data['user']['userNo'],
                    "user_name": global_data['user']['name'],
                    "course_code": global_data['course']['courseCode'],
                    "course_name": global_data['course']['name'],
                    "dep_id": global_data['dept']['id'],
                    "dep_name": global_data['dept']['name'],
                    "dep_code": global_data['dept']['code'],
                }
            )
            if rsp.status_code != 204:
                raise ValueError(f"Failed to read video chunk: {rsp.json()}")

        duration = 0
        if len(self.uploads) > 0:
            duration = self.uploads[0].videos[0].duration
        elif self._raw_data['data'].get('duration') is not None:
            duration = self._raw_data['data']['duration']
        else:
            raise ValueError("Failed to get video duration")

        duration = int(duration)

        start = 0
        step = 120
        while start < duration:
            end = min(start + step, duration)
            _read_chunk(start, end)
            start += step

    def _complete_material(self):
        for upload in self.uploads:
            self._session.post(
                url=f"https://eclass.yuntech.edu.tw/api/course/activities-read/{self.id}",
                json={"upload_id": upload.id},
            )
            # if response.ok:
            #     print(f"已自動閱讀 {upload.name}")
            # else:
            #     print(f"自動閱讀 {upload.name} 失敗")
            #     print(response.json())
            #     print(response.status_code)

    def complete(self):
        """ 完成活動指標 """
        if self.type == 'online_video':
            self._complete_video()
        elif self.type == 'material':
            self._complete_material()
        else:
            raise ValueError(f"Unknown activity type: {self.type}")



class CaptchaError(Exception):
    """Exception raised for errors related to captcha verification."""

    def __init__(self, message="Captcha verification failed"):
        self.message = message
        super().__init__(self.message)


class AccessDenyError(Exception):
    """Exception raised for errors related to access denied."""

    def __init__(self, message="Access denied"):
        self.message = message
        super().__init__(self.message)


class AccountNotRegisteredError(Exception):
    """Exception raised for errors in the account existence or registration status."""

    def __init__(self, message="Account not exist or registered."):
        self.message = message
        super().__init__(self.message)



import json
import re

from bs4 import BeautifulSoup
from requests import Session


def get_global_data(session: Session, url) -> dict:
    response = session.get(url)
    html_content = response.text

    # Step 2: 使用 BeautifulSoup 解析 HTML
    soup = BeautifulSoup(html_content, 'html.parser')
    script_tags = soup.find_all('script', text=True)
    for script in script_tags:
        if 'var globalData =' in script.string:
            script_content = script.string
            break
    else:
        raise Exception("Unable to find globalData script")

    # Step 3: 使用正則表達式提取 globalData JSON
    match = re.search(r'var globalData\s*=\s*({.*})', script_content, re.DOTALL)
    if not match:
        raise Exception("globalData not found in script")

    global_data_raw = match.group(1)

    def parse_global_data_without_quotes(script_content):
        # 使用正則處理，將 key 補上雙引號
        script_content = re.sub(r'(\b[a-zA-Z0-9_]+\b)\s*:', r'"\1":', script_content)
        # 移除多餘的逗號：在 JSON 的 } 或 ] 之前出現的多餘逗號
        script_content = re.sub(r',(\s*[}\]])', r'\1', script_content)
        return script_content

    # 轉換為合法 JSON 格式
    global_data_json = parse_global_data_without_quotes(global_data_raw)

    return json.loads(global_data_json)



from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

from ..utils import get_global_data
from .activity import Activity
from .base import BaseResourceModel


class Department(BaseModel):
    id: int
    name: str


class Course(BaseResourceModel):
    id: int
    name: str
    department: Department | None

    @property
    def global_data(self) -> dict:
        if self._global_data is None:
            self._global_data = get_global_data(
                self._session,
                f"https://eclass.yuntech.edu.tw/course/{self.id}/learning-activity/full-screen"
            )
        return self._global_data

    def __init__(self, *, session, **data):
        super().__init__(session=session, **data)
        self._global_data = None

    def fetch_coursewares(self) -> list["Activity"]:
        response = self._session.get(f"https://eclass.yuntech.edu.tw/api/course/{self.id}/coursewares?page_size=1000")
        data = response.json()['activities']

        return [Activity(session=self._session, course=self, **d) for d in data]



from typing import Any

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr
from requests import Session


class BaseResourceModel(BaseModel):
    _raw_data: dict[str, Any] = PrivateAttr(default_factory=dict)

    def __init__(self, *, session: Session, **data):
        super().__init__(**data)
        self._raw_data = data
        self._session = session

    model_config = ConfigDict(extra="allow")  # 允許額外的欄位



import json
import logging
import re
import time

import requests
from bs4 import BeautifulSoup
from colorlog import ColoredFormatter

from .entity.activity import Activity
from .entity.course import Course


class TronClassClient:
    def __init__(self, session: requests.Session, logger=None) -> None:
        self.session = session
        self.base_url = "https://eclass.yuntech.edu.tw"
        if logger:
            self.logger = logger
        else:
            self.logger = self._create_default_logger()

    def _create_default_logger(self):
        logger = logging.getLogger('TronClassClient')
        logger.setLevel(logging.DEBUG)

        formatter = ColoredFormatter(
            "%(log_color)s%(levelname)s:%(reset)s [%(name)s] %(message)s",
            datefmt='%Y-%m-%d %H:%M:%S',
            log_colors={
                "DEBUG": "white",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "bold_red",
            },
        )

        handler = logging.StreamHandler()
        handler.setFormatter(formatter)

        logger.addHandler(handler)
        return logger

    def _get(self, url: str) -> dict:
        response = self.session.get(f"{self.base_url}{url}")
        return response.json()

    def get_global_data(self, url) -> dict:
        response = self.session.get(url)
        html_content = response.text

        # Step 2: 使用 BeautifulSoup 解析 HTML
        soup = BeautifulSoup(html_content, 'html.parser')
        script_tags = soup.find_all('script', text=True)
        for script in script_tags:
            if 'var globalData =' in script.string:
                script_content = script.string
                break
        else:
            raise Exception("Unable to find globalData script")

        # Step 3: 使用正則表達式提取 globalData JSON
        match = re.search(r'var globalData\s*=\s*({.*})', script_content, re.DOTALL)
        if not match:
            raise Exception("globalData not found in script")

        global_data_raw = match.group(1)

        def parse_global_data_without_quotes(script_content):
            # 使用正則處理，將 key 補上雙引號
            script_content = re.sub(r'(\b[a-zA-Z0-9_]+\b)\s*:', r'"\1":', script_content)
            # 移除多餘的逗號：在 JSON 的 } 或 ] 之前出現的多餘逗號
            script_content = re.sub(r',(\s*[}\]])', r'\1', script_content)
            return script_content

        # 轉換為合法 JSON 格式
        global_data_json = parse_global_data_without_quotes(global_data_raw)

        return json.loads(global_data_json)

    def fetch_courses(self) -> list[Course]:
        courses_data = self._get("/api/my-courses?page_size=1000")['courses']
        return [Course(session=self.session, **course_data) for course_data in courses_data]

    def fetch_activities(self, course_id: int) -> list[Activity]:
        activities_data = self._get(f"/api/course/{course_id}/coursewares?page_size=1000")['activities']
        return [Activity(**activity_data) for activity_data in activities_data]

    def fetch_activity(self, activity_id: int) -> Activity:
        activity_data = self._get(f"/api/activities/{activity_id}")
        return Activity(**activity_data)

    def _read_chunk_video(self, course_id: int, activity_id: int, activity_data: dict, start: int, end: int):
        global_data = self.get_global_data(f"https://eclass.yuntech.edu.tw/course/{course_id}/learning-activity/full-screen")

        response = self.session.post(
            f"https://eclass.yuntech.edu.tw/api/course/activities-read/{activity_id}",
            json={"start": start, "end": end}
        )

        response = self.session.post(
            "https://eclass.yuntech.edu.tw/statistics/api/online-videos",
            json={
                "user_id": global_data['user']['id'],
                "org_id": global_data['course']['orgId'],
                "course_id": global_data['course']['id'],
                "module_id": activity_data['module_id'],
                "syllabus_id": activity_data['syllabus_id'],
                "activity_id": activity_id,
                "upload_id": activity_data['uploads'][0]['id'],
                "reply_id": None,
                "comment_id": None,
                "forum_type": "",
                "action_type": "play",
                "is_teacher": False,
                "is_student": True,
                "ts": int(time.time() * 1000),
                "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36 Edg/131.0.0.0",
                "meeting_type": "online_video",
                "start_at": start,
                "end_at": end,
                "duration": end - start,
                "org_name": global_data['course']['orgName'],
                "org_code": global_data['course']['orgCode'],
                "user_no": global_data['user']['userNo'],
                "user_name": global_data['user']['name'],
                "course_code": global_data['course']['courseCode'],
                "course_name": global_data['course']['name'],
                "dep_id": global_data['dept']['id'],
                "dep_name": global_data['dept']['name'],
                "dep_code": global_data['dept']['code'],
            }
        )

    def _read_video_activity(self, activity_id: int):
        self.logger.info(f"Reading full online video: {activity_id}")
        activity_data = self._get(f"/api/activities/{activity_id}")
        if activity_data['type'] != 'online_video':
            raise ValueError(f"Activity {activity_id} is not an online video")

        duration = int(activity_data['uploads'][0]['videos'][0]['duration'])

        start = 0
        step = 120
        while start < duration:
            end = min(start + step, duration)
            self._read_chunk_video(activity_data['course_id'], activity_id, activity_data, start, end)
            start += step
